

         /*********************************************************/
         /*                                                       */
         /*                 COPYRIGHT (C) 1988 BY                 */
         /* CAD Group of Dipartimento di Automatica e Informatica */
         /*             Politecnico di Torino, Italy.             */
         /*                  ALL RIGHTS RESERVED.                 */
         /*                                                       */
         /* This software is furnished under a license and may be */
         /* used and copied  only in accordance with the terms of */
         /* such license. Unauthorized copies will be persecuted. */
         /*                                                       */
         /*********************************************************/





/******************************************************************************

    Filename:         tohilo.c

    Author:           Pier Luca Montessoro

    Creation date:    21 June 1988

    Contents:         flat translator to hilo hdl


                          *** DOCUMENTATION ***

 MAIN

      is the main function.


 LOAD_TOPOLOGY

      loads the topological information.


 TRANSLATE

      writes the translation on the output file.


 GOOD_BYE

      is the exit function.


 GET_MEMORY

      allocs the memory needed for the topological data structures.


 LOAD_MAIN_SYM_TAB

      loads the symbol table of the main unit.


 ******************************************************************************/



#include "env.h"
#include "loadcode.h"


#define  LINKER_VERSION     "3.0"

#define  MAX_FILE_NAME_LENGTH  64
#define  MAX_NAME_LENGTH       32
#define  MAX_MESSAGE_LENGTH   132
#define  MAX_DATA_SIZE        112
#define  BUFFER_SIZE           64

#define  NET_EXTENSION      "dee"
#define  HILO_EXTENSION     "hil"

#define  RISE_DEL_COEFF        1
#define  FALL_DEL_COEFF        1



/* maximum delays */

unsigned long  max_r_del = 0;
unsigned long  max_f_del = 0;


/* delay table element structure definition */

typedef struct Delay_Table_Element
        {
           unsigned long  rise;
           unsigned long  fall;
        }  DELAY_TABLE_ELEMENT;


/* fanin and fanout structure definitions */

typedef struct Fanin_Element
        {
           unsigned long  index;
        }  FANIN_ELEMENT;

typedef struct Fanout_Element
        {
           unsigned long  index;
        }  FANOUT_ELEMENT;


/* topological element structure definition */

typedef struct Topological_Element
        {
           unsigned char   it_is_in;
           unsigned char   it_is_out;
           unsigned char   it_is_only_a_wire;
           unsigned char   unused;
           unsigned long   delay_index;
           unsigned long   fanin_position;
           unsigned long   fanin_number;
           unsigned long   fanout_position;
           unsigned long   fanout_number;
           char            name [MAX_NAME_LENGTH];
           char            type [MAX_NAME_LENGTH];

        }  TOPOLOGICAL_ELEMENT;


/* delay table declaration */

DELAY_TABLE_ELEMENT    *delay_table;


/* topological vectors */

TOPOLOGICAL_ELEMENT    *topology_vector;
FANIN_ELEMENT          *fanin_vector;
FANOUT_ELEMENT         *fanout_vector;


/* number of elements */

unsigned long          elements_num;


/* experiment name */

char                    experiment_name [MAX_NAME_LENGTH];


/* input topological file */

FILE                   *net_file;


/* output topological file */

FILE                   *hilo_file;



/******************************************************************************

    Function name:    main

    Authors:          Pier Luca Montessoro

    Creation date:    21 June 1988

    Purpose:          main

 ******************************************************************************/


main (argc, argv)

int   argc;
char  *argv[];

{

   /* local variables */

   char                  net_file_name [MAX_FILE_NAME_LENGTH];
   char                  hilo_file_name [MAX_FILE_NAME_LENGTH];
   unsigned char         comp_delay;


   /* beginning of executable part */

   if (argc != 3)
   {
      printf 
      ("Wrong number of arguments (1st: netfile, 2nd: d/n - delay/nodelay).\n");
      good_bye (FALSE, TRUE);
   }

   if ((argv[2][0] == 'n') || (argv[2][0] == 'N'))
   {
      comp_delay = FALSE;
   }
   else
   {    
      comp_delay = TRUE;
   }


   /* open the network and the hilo files */

   strcpy (net_file_name, argv[1]);
   strcat (net_file_name, ".");
   strcat (net_file_name, NET_EXTENSION);

   strcpy (hilo_file_name, argv[1]);
   strcat (hilo_file_name, ".");
   strcat (hilo_file_name, HILO_EXTENSION);

   if ((net_file = fopen (net_file_name, "r")) == NULL)
   {
      printf ("I cannot open the network file.\n");
      good_bye (FALSE, TRUE);
   }

   if ((hilo_file = fopen (hilo_file_name, "w")) == NULL)
   {
      printf ("I cannot open the output file.\n");
      good_bye (TRUE, TRUE);
   }

   load_topology();
   translate (comp_delay);

   good_bye (TRUE, FALSE);

}  /* end of main */



/******************************************************************************

    Function name:    load_topology

    Authors:          Pier Luca Montessoro

    Creation date:    21 June 1988

    Purpose:          to load the topological information

 ******************************************************************************/


load_topology ()

{

   /* called function */

   char           *get_memory();


   /* local variables */

   char              out_buffer [BUFFER_SIZE];
   unsigned long     input_counter = 0;
   unsigned long     output_counter = 0;


   /* general information variables */

   char              linker_version [MAX_NAME_LENGTH];
   char              circuit_name [MAX_NAME_LENGTH];


   /* symbol table and module table offsets */

   long int          sym_tab_offset;
   long int          mod_tab_offset;


   /* number of warnings */

   unsigned long     warnings;


   /* information related to the topological elements */

   unsigned long     p_o_num;
   unsigned long     p_i_num;
   unsigned long     p_b_num;
   unsigned long     p_o_size;
   unsigned long     p_i_size;
   unsigned long     p_b_size;
   unsigned long     constants_num;
   unsigned long     nodes_num;
   unsigned long     wires_num;
   unsigned long     regi_num;
   unsigned long     wires_size;
   unsigned long     regi_size;
   unsigned long     transistors_num;
   unsigned long     gates_num;
   unsigned long     rt_assignement_num;
   unsigned long     rt_condition_num;
   unsigned long     memory_num;
   unsigned long     locvar_num;
   unsigned long     proc_num;
   unsigned long     unused;
   unsigned long     gates_array_num;

   /* delays */

   unsigned long     delays_num;
   unsigned long     delay_unit;
   unsigned long     min_delay;
   unsigned long     max_delay;

   /* initialization table */

   unsigned long     init_num;

   /* C-modules and functions */

   unsigned long     c_modules_num;
   unsigned long     c_functions_num;

   /* maximum parameters */

   unsigned long     max_strength;
   unsigned long     max_size;
   unsigned long     max_fanout;
   unsigned long     max_atpg_level;

   /* link date and time */

   unsigned long     link_date;
   unsigned long     link_time;
   unsigned long     c_link_date;

   /* MOZART internal information */

   unsigned long     vdd_offset;
   unsigned long     gnd_offset;
   unsigned long     static_size;

   /* flags */

   unsigned long     atpg_flag;
   unsigned long     sim_flag;
   unsigned long     gate_array_flag;
   unsigned long     case_sensitive_flag;

   /* topological data structure sizes */

   unsigned long     tot_elements_num;
   unsigned long     tot_fanin_num;
   unsigned long     tot_fanout_num;
   unsigned long     tot_bidir_num;

   /* averages values */

   unsigned long     nodes_fo;
   unsigned long     nodes_fi;
   unsigned long     nodes_conn;
   unsigned long     wires_fo;
   unsigned long     wires_fi;
   unsigned long     regi_fo;
   unsigned long     regi_fi;
   unsigned long     gates_fo;
   unsigned long     gates_fi;
   unsigned long     rt_a_fo;
   unsigned long     rt_a_fi;
   unsigned long     rt_c_fo;
   unsigned long     rt_c_fi;



   /* beginning of executable part */

   /* read the header information from the network file */

   fscanf (net_file, "%s", linker_version);

   /* check the linker version */

   if (strcmp (linker_version, LINKER_VERSION))
   {
      printf ("The linker version is not compatible with me.\n");
      good_bye (TRUE, TRUE);
   }

   fscanf (net_file, "%s", circuit_name);

   fscanf (net_file, "%ld", &sym_tab_offset);
   fscanf (net_file, "%ld", &mod_tab_offset);

   fscanf (net_file, "%ld", &warnings);

   fscanf (net_file, "%ld", &p_o_num);
   fscanf (net_file, "%ld", &p_o_size);
   fscanf (net_file, "%ld", &p_i_num);
   fscanf (net_file, "%ld", &p_i_size);
   fscanf (net_file, "%ld", &p_b_num);
   fscanf (net_file, "%ld", &p_b_size);
   fscanf (net_file, "%ld", &constants_num);
   fscanf (net_file, "%ld", &nodes_num);
   fscanf (net_file, "%ld", &wires_num);
   fscanf (net_file, "%ld", &wires_size);
   fscanf (net_file, "%ld", &regi_num);
   fscanf (net_file, "%ld", &regi_size);
   fscanf (net_file, "%ld", &transistors_num);
   fscanf (net_file, "%ld", &gates_num);
   fscanf (net_file, "%ld", &rt_assignement_num);
   fscanf (net_file, "%ld", &rt_condition_num);
   fscanf (net_file, "%ld", &memory_num);
   fscanf (net_file, "%ld", &locvar_num);
   fscanf (net_file, "%ld", &proc_num);
   fscanf (net_file, "%ld", &unused);
   fscanf (net_file, "%ld", &gates_array_num);

   fscanf (net_file, "%ld", &delays_num);
   fscanf (net_file, "%ld", &delay_unit);
   fscanf (net_file, "%ld", &min_delay);
   fscanf (net_file, "%ld", &max_delay);

   fscanf (net_file, "%ld", &init_num);

   fscanf (net_file, "%ld", &unused);
   fscanf (net_file, "%ld", &unused);

   fscanf (net_file, "%ld", &c_modules_num);
   fscanf (net_file, "%ld", &c_functions_num);

   fscanf (net_file, "%ld", &max_strength);
   fscanf (net_file, "%ld", &max_size);
   fscanf (net_file, "%ld", &max_fanout);
   fscanf (net_file, "%ld", &max_atpg_level);

   fscanf (net_file, "%ld", &vdd_offset);
   fscanf (net_file, "%ld", &gnd_offset);

   fscanf (net_file, "%ld", &link_date);
   fscanf (net_file, "%ld", &link_time);
   fscanf (net_file, "%ld", &c_link_date);

   fscanf (net_file, "%ld", &static_size);

   fscanf (net_file, "%ld", &atpg_flag);
   fscanf (net_file, "%ld", &sim_flag);
   fscanf (net_file, "%ld", &gate_array_flag);
   fscanf (net_file, "%ld", &case_sensitive_flag);

   fscanf (net_file, "%ld", &tot_elements_num);
   fscanf (net_file, "%ld", &tot_fanin_num);
   fscanf (net_file, "%ld", &tot_fanout_num);
   fscanf (net_file, "%ld", &tot_bidir_num);

   fscanf (net_file, "%ld", &nodes_fo);
   fscanf (net_file, "%ld", &nodes_fi);
   fscanf (net_file, "%ld", &nodes_conn);
   fscanf (net_file, "%ld", &wires_fo);
   fscanf (net_file, "%ld", &wires_fi);
   fscanf (net_file, "%ld", &regi_fo);
   fscanf (net_file, "%ld", &regi_fi);
   fscanf (net_file, "%ld", &gates_fo);
   fscanf (net_file, "%ld", &gates_fi);
   fscanf (net_file, "%ld", &rt_a_fo);
   fscanf (net_file, "%ld", &rt_a_fi);
   fscanf (net_file, "%ld", &rt_c_fo);
   fscanf (net_file, "%ld", &rt_c_fi);



   /* save the circuit name */

   strcpy (experiment_name, circuit_name);


   /* allocate the delay table and read the delay values */

   delay_table = (DELAY_TABLE_ELEMENT *)
                      get_memory ((unsigned long) delays_num,
                                  (unsigned long) sizeof (DELAY_TABLE_ELEMENT));

   {

      /* local variable */

      unsigned long  delay_counter;
      unsigned long  fall_tmp;
      unsigned long  rise_tmp;


      for (delay_counter = 0; delay_counter < delays_num; delay_counter++)
      {
         fscanf (net_file, "%ld %ld", &rise_tmp, &fall_tmp);
         delay_table[delay_counter].rise = rise_tmp;
         delay_table[delay_counter].fall = fall_tmp;
      }
   }



   /* the C-module section, C-module table and C-function table are
      not yet implemeted */

   if (c_modules_num)
   {
      printf ("C modules are not supported.\n");
      good_bye (TRUE, TRUE);
   }

   if (c_functions_num)
   {
      printf ("C functions are not supported.\n");
      good_bye (TRUE, TRUE);
   }



   /* allocate the memory for topological data structures */

   elements_num = tot_elements_num;

   topology_vector = (TOPOLOGICAL_ELEMENT *) get_memory
                             ((unsigned long) tot_elements_num,
                              (unsigned long) sizeof (TOPOLOGICAL_ELEMENT));

   /* clear the name fields */

   {
      unsigned long  i;

      for (i = 0; i < tot_elements_num; i++)
      {
         *topology_vector[i].name = '\0';
      }
   }


   /* now the names of the elements in the main unit can be loaded */

   load_sym_tab();


   if (tot_fanin_num > 0)
      fanin_vector = (FANIN_ELEMENT *) get_memory
                             ((unsigned long) tot_fanin_num,
                              (unsigned long) sizeof (FANIN_ELEMENT));

   if (tot_fanout_num > 0)
      fanout_vector = (FANOUT_ELEMENT *) get_memory
                             ((unsigned long) tot_fanout_num,
                              (unsigned long) sizeof (FANOUT_ELEMENT));

   if (tot_bidir_num > 0)
   {
      printf ("The connectivity lists are not supported.\n");
      good_bye (TRUE, TRUE);
   }


   /* load the topological elements */

   {

      /* local variables */

      unsigned long   te_counter;
      unsigned long   te_attribute;
      unsigned long   te_code;
      unsigned long   te_sim_level;
      unsigned long   te_atpg_level;
      unsigned long   te_dim;
      unsigned long   te_size;
      unsigned long   te_type;
      unsigned long   te_strength;
      unsigned long   te_0_strength;
      unsigned long   te_1_strength;
      unsigned long   te_delay_index;
      unsigned long   te_fo_num;
      unsigned long   te_fi_num;
      unsigned long   te_conn_num;
      unsigned long   fo_offset;
      unsigned long   fi_offset;
      unsigned long   link_counter;
      unsigned long   link_type;
      unsigned long   link_index;
      unsigned long   link_bit;
      unsigned long   link_b1c;
      unsigned long   link_b2c;
      unsigned long   link_b1d;
      unsigned long   link_b2d;
      unsigned long   link_item;
      unsigned long   link_pin;
      unsigned char   const_string [MAX_DATA_SIZE];


      /* beginning of executable part */

      fo_offset = fi_offset = 0;

      for (te_counter = 0; te_counter < tot_elements_num; te_counter++)
      {

         /* read the first record of the topological descriptor and load some
            fields */

         fscanf (net_file, "%ld %ld %ld %ld", &te_attribute, &te_code,
                                              &te_sim_level, &te_atpg_level);

         switch ( (unsigned char) te_attribute)
         {
            case MZ_INTERNAL      : 
                 sprintf (topology_vector[te_counter].name, 
                          "wx%d", te_counter);
                 topology_vector[te_counter].it_is_in = FALSE;
                 topology_vector[te_counter].it_is_out = FALSE;
                 break;

            case MZ_INPUT         : 
                 if ((te_counter != gnd_offset) && (te_counter != vdd_offset))
                 {
                    /* check if the name has been loaded */
                    if (*topology_vector[te_counter].name == '\0')
                    {
                       printf 
     ("Element #%ld has been recognized as input, but its name is unknown.\n",
                       te_counter);
                       good_bye (TRUE, TRUE);
                    }
                    topology_vector[te_counter].it_is_in = TRUE;
                    topology_vector[te_counter].it_is_out = FALSE;
                 }
                 else
                 {
                    if (te_counter == gnd_offset)
                    {
                       strcpy (topology_vector[te_counter].name, "GND");
                    }
                    if (te_counter == vdd_offset)
                    {
                       strcpy (topology_vector[te_counter].name, "VDD");
                    }

                    topology_vector[te_counter].it_is_in = FALSE;
                    topology_vector[te_counter].it_is_out = FALSE;
                 }
                 break;

            case MZ_OUTPUT        : 
                 /* check if the name has been loaded */
                 if (*topology_vector[te_counter].name == '\0')
                 {
                    printf 
     ("Element #%ld has been recognized as output, but its name is unknown.\n",
                    te_counter);
                    good_bye (TRUE, TRUE);
                 }
                 topology_vector[te_counter].it_is_in = FALSE;
                 topology_vector[te_counter].it_is_out = TRUE;
                 break;

            case MZ_BIDIRECTIONAL : printf 
                 ("The bidirectional elements are not supported.\n");
                 good_bye (FALSE, TRUE);
                 break;

            case MZ_PHANTOM       : 
                 sprintf (topology_vector [te_counter].name, 
                          "wx%d", te_counter);
                 topology_vector[te_counter].it_is_in = FALSE;
                 topology_vector[te_counter].it_is_out = FALSE;
                 break;
         }


         switch ( (unsigned char) te_code)
         {
            case MZ_NODE:

                     /* read the second record of the topological descriptor */

                     fscanf (net_file, "%ld %ld %ld %ld %ld %ld",
                                       &te_dim, &te_size, &te_delay_index,
                                       &te_fo_num, &te_fi_num, &te_conn_num);

                     if (te_delay_index != 0)
                     {
                        printf 
                           ("WARNING: a delay on node has been ignored.\n"); 
                     }

                     topology_vector[te_counter].it_is_only_a_wire = TRUE;

                     topology_vector[te_counter].fanin_position = fi_offset;
                     topology_vector[te_counter].fanout_position = fo_offset;
                     topology_vector[te_counter].fanin_number = te_fi_num;
                     topology_vector[te_counter].fanout_number = te_fo_num;

                     if (te_dim != 1)
                     {
                        printf ("Vectorial nodes are not supported.\n");
                        good_bye (TRUE, TRUE);
                     }

                     break;

            case MZ_WIRE:

                     /* read the second record of the topological descriptor */

                     fscanf (net_file, "%ld %ld %ld %ld %ld %ld",
                                       &te_dim, &te_size, &te_delay_index,
                                       &te_fo_num, &te_fi_num, &te_conn_num);

                     topology_vector[te_counter].it_is_only_a_wire = TRUE;

                     topology_vector[te_counter].fanin_position = fi_offset;
                     topology_vector[te_counter].fanout_position = fo_offset;
                     topology_vector[te_counter].fanin_number = te_fi_num;
                     topology_vector[te_counter].fanout_number = te_fo_num;

                     if (te_delay_index != 0)
                     {
                        printf 
                           ("WARNING: a delay on wire has been ignored.\n"); 
                     }

                     if (te_dim > MAX_DATA_SIZE)
                     {
                        printf 
                           ("Unlimited data bit-length is not supported.\n");
                        good_bye (TRUE, TRUE);
                     }

                     break;

            case MZ_CONST:

                     /* read the second record of the topological descriptor */

                     fscanf (net_file, "%s %ld %ld", const_string,
                                                     &te_fo_num, &te_conn_num);

                     printf ("I don't know how to translate constants.\n");
                     good_bye (TRUE, TRUE);

                     topology_vector[te_counter].it_is_only_a_wire = TRUE;

                     topology_vector[te_counter].fanin_position = fi_offset;
                     topology_vector[te_counter].fanout_position = fo_offset;
                     topology_vector[te_counter].fanin_number = 0;
                     topology_vector[te_counter].fanout_number = te_fo_num;

                     if (te_dim > MAX_DATA_SIZE)
                     {
                        printf 
                           ("Unlimited data bit-length is not supported.\n");
                        good_bye (TRUE, TRUE);
                     }

                     break;

            case MZ_TRANSISTOR:

                     /* read the second record of the topological descriptor */

                     fscanf (net_file, "%ld %ld", &te_type, &te_strength);

                     printf ("I don't know how to translate transistors.\n");
                     good_bye (TRUE, TRUE);

                     topology_vector[te_counter].it_is_only_a_wire = FALSE;

                     topology_vector[te_counter].fanin_position = fi_offset;
                     topology_vector[te_counter].fanout_position = fo_offset;
                     topology_vector[te_counter].fanin_number = 0;
                     topology_vector[te_counter].fanout_number = 0;

                     switch ( (unsigned char) te_type)
                     {
                        case MZ_DEPLETION:
                               break;
                        case MZ_N_CHANNEL:
                               break;
                        case MZ_P_CHANNEL:
                               break;
                        default: 
                               printf ("I found an unknown transistor type.\n");
                               good_bye (TRUE, TRUE);
                     }

                     break;

            case MZ_GATE:

                     /* read the second record of the topological descriptor */

                     fscanf (net_file, "%ld %ld %ld %ld %ld %ld",
                                      &te_type, &te_delay_index, &te_0_strength,
                                      &te_1_strength, &te_fo_num, &te_fi_num);

                     topology_vector[te_counter].delay_index = te_delay_index;

                     topology_vector[te_counter].it_is_only_a_wire = FALSE;

                     topology_vector[te_counter].fanin_position = fi_offset;
                     topology_vector[te_counter].fanout_position = fo_offset;
                     topology_vector[te_counter].fanin_number = te_fi_num;
                     topology_vector[te_counter].fanout_number = te_fo_num;

                     switch ( (unsigned char) te_type)
                     {
                        case MZ_AND       :
                                 strcpy (topology_vector[te_counter].type, 
                                          "AND");
                                 break;
                        case MZ_NAND      :
                                 strcpy (topology_vector[te_counter].type, 
                                          "NAND");
                                 break;
                        case MZ_OR        :
                                 strcpy (topology_vector[te_counter].type, 
                                          "OR");
                                 break;
                        case MZ_NOR       :
                                 strcpy (topology_vector[te_counter].type, 
                                          "NOR");
                                 break;
                        case MZ_BUF       :
                                 strcpy (topology_vector[te_counter].type, 
                                          "BUF");
                                 break;
                        case MZ_INV       :
                                 strcpy (topology_vector[te_counter].type, 
                                          "NOT");
                                 break;
                        case MZ_XOR       :
#if 0
                                 strcpy (topology_vector[te_counter].type, 
                                          "XOR");
#endif
#if 1
                                 printf ("I cannot translate %s element.\n",
                                          "MZ_XOR");
                                 good_bye (TRUE, TRUE);
#endif
                                 break;
                        case MZ_XNOR      :
#if 0
                                 strcpy (topology_vector[te_counter].type, 
                                          "XNOR");
#endif
#if 1
                                 printf ("I cannot translate %s element.\n",
                                          "MZ_XNOR");
                                 good_bye (TRUE, TRUE);
#endif
                                 break;
                        case MZ_TGP       :
                                 strcpy (topology_vector[te_counter].type, 
                                          "MOVEIF1");
                                 break;
                        case MZ_TGN       :
                                 strcpy (topology_vector[te_counter].type, 
                                          "MOVEIF0");
                                 break;
                        case MZ_MTGP      :
#if 0
                                 strcpy (topology_vector[te_counter].type, 
                                          "MZ_MTGP");
#endif
                                 printf ("I cannot translate %s element.\n",
                                          "MZ_MTGP");
                                 good_bye (TRUE, TRUE);
                                 break;
                        case MZ_MTGN      :
#if 0
                                 strcpy (topology_vector[te_counter].type, 
                                          "MZ_MTGN");
#endif
                                 printf ("I cannot translate %s element.\n",
                                          "MZ_MTGN");
                                 good_bye (TRUE, TRUE);
                                 break;
                        case MZ_BUS_TRI   :
#if 0
                                 strcpy (topology_vector[te_counter].type, 
                                          "MZ_BUS_TRI");
#endif
                                 printf ("I cannot translate %s element.\n",
                                          "MZ_BUS_TRI");
                                 good_bye (TRUE, TRUE);
                                 break;
                        case MZ_BUS_MOS   :
#if 0
                                 strcpy (topology_vector[te_counter].type, 
                                          "MZ_BUS_MOS");
#endif
                                 printf ("I cannot translate %s element.\n",
                                          "MZ_BUS_MOS");
                                 good_bye (TRUE, TRUE);
                                 break;
                        case MZ_BUS_UP    :
#if 0
                                 strcpy (topology_vector[te_counter].type, 
                                          "MZ_BUS_UP");
#endif
                                 printf ("I cannot translate %s element.\n",
                                          "MZ_BUS_UP");
                                 good_bye (TRUE, TRUE);
                                 break;
                        case MZ_BUS_DOWN  :
#if 0
                                 strcpy (topology_vector[te_counter].type, 
                                          "MZ_BUS_DOWN");
#endif
                                 printf ("I cannot translate %s element.\n",
                                          "MZ_BUS_DOWN");
                                 good_bye (TRUE, TRUE);
                                 break;
                        case MZ_AND_TRI   :
#if 0
                                 strcpy (topology_vector[te_counter].type, 
                                          "MZ_AND_TRI");
#endif
                                 printf ("I cannot translate %s element.\n",
                                          "MZ_AND_TRI");
                                 good_bye (TRUE, TRUE);
                                 break;
                        case MZ_NAND_TRI  :
#if 0
                                 strcpy (topology_vector[te_counter].type, 
                                          "MZ_NAND_TRI");
#endif
                                 printf ("I cannot translate %s element.\n",
                                          "MZ_NAND_TRI");
                                 good_bye (TRUE, TRUE);
                                 break;
                        case MZ_OR_TRI    :
#if 0
                                 strcpy (topology_vector[te_counter].type, 
                                          "MZ_OR_TRI");
#endif
                                 printf ("I cannot translate %s element.\n",
                                          "MZ_OR_TRI");
                                 good_bye (TRUE, TRUE);
                                 break;
                        case MZ_NOR_TRI   :
#if 0
                                 strcpy (topology_vector[te_counter].type, 
                                          "MZ_NOR_TRI");
#endif
                                 printf ("I cannot translate %s element.\n",
                                          "MZ_NOR_TRI");
                                 good_bye (TRUE, TRUE);
                                 break;
                        case MZ_BUF_TRI   :
                                 strcpy (topology_vector[te_counter].type, 
                                          "BUFIF1");
                                 break;
                        case MZ_INV_TRI   :
                                 strcpy (topology_vector[te_counter].type, 
                                          "NOTIF1");
                                 break;
                        case MZ_AND_TRIN  :
#if 0
                                 strcpy (topology_vector[te_counter].type, 
                                          "MZ_AND_TRIN");
#endif
                                 printf ("I cannot translate %s element.\n",
                                          "MZ_AND_TRIN");
                                 good_bye (TRUE, TRUE);
                                 break;
                        case MZ_NAND_TRIN :
#if 0
                                 strcpy (topology_vector[te_counter].type, 
                                          "MZ_NAND_TRIN");
#endif
                                 printf ("I cannot translate %s element.\n",
                                          "MZ_NAND_TRIN");
                                 good_bye (TRUE, TRUE);
                                 break;
                        case MZ_OR_TRIN   :
#if 0
                                 strcpy (topology_vector[te_counter].type, 
                                          "MZ_OR_TRIN");
#endif
                                 printf ("I cannot translate %s element.\n",
                                          "MZ_OR_TRIN");
                                 good_bye (TRUE, TRUE);
                                 break;
                        case MZ_NOR_TRIN  :
#if 0
                                 strcpy (topology_vector[te_counter].type, 
                                          "MZ_NOR_TRIN");
#endif
                                 printf ("I cannot translate %s element.\n",
                                          "MZ_NOR_TRIN");
                                 good_bye (TRUE, TRUE);
                                 break;
                        case MZ_BUF_TRIN  :
                                 strcpy (topology_vector[te_counter].type, 
                                          "BUFIF0");
                                 break;
                        case MZ_INV_TRIN  :
                                 strcpy (topology_vector[te_counter].type, 
                                          "NOTIF0");
                                 break;
                        case MZ_AND_OC    :
#if 0
                                 strcpy (topology_vector[te_counter].type, 
                                          "MZ_AND_OC");
#endif
                                 printf ("I cannot translate %s element.\n",
                                          "MZ_AND_OC");
                                 good_bye (TRUE, TRUE);
                                 break;
                        case MZ_NAND_OC   :
#if 0
                                 strcpy (topology_vector[te_counter].type, 
                                          "MZ_NAND_OC");
#endif
                                 printf ("I cannot translate %s element.\n",
                                          "MZ_NAND_OC");
                                 good_bye (TRUE, TRUE);
                                 break;
                        case MZ_OR_OC     :
#if 0
                                 strcpy (topology_vector[te_counter].type, 
                                          "MZ_OR_OC");
#endif
                                 printf ("I cannot translate %s element.\n",
                                          "MZ_OR_OC");
                                 good_bye (TRUE, TRUE);
                                 break;
                        case MZ_NOR_OC    :
#if 0
                                 strcpy (topology_vector[te_counter].type, 
                                          "MZ_NOR_OC");
#endif
                                 printf ("I cannot translate %s element.\n",
                                          "MZ_NOR_OC");
                                 good_bye (TRUE, TRUE);
                                 break;
                        case MZ_BUF_OC    :
#if 0
                                 strcpy (topology_vector[te_counter].type, 
                                          "MZ_BUF_OC");
#endif
                                 printf ("I cannot translate %s element.\n",
                                          "MZ_BUF_OC");
                                 good_bye (TRUE, TRUE);
                                 break;
                        case MZ_INV_OC    :
#if 0
                                 strcpy (topology_vector[te_counter].type, 
                                          "MZ_INV_OC");
#endif
                                 printf ("I cannot translate %s element.\n",
                                          "MZ_INV_OC");
                                 good_bye (TRUE, TRUE);
                                 break;
                        case MZ_AND_OE    :
#if 0
                                 strcpy (topology_vector[te_counter].type, 
                                          "MZ_AND_OE");
#endif
                                 printf ("I cannot translate %s element.\n",
                                          "MZ_AND_OE");
                                 good_bye (TRUE, TRUE);
                                 break;
                        case MZ_NAND_OE   :
#if 0
                                 strcpy (topology_vector[te_counter].type, 
                                          "MZ_NAND_OE");
#endif
                                 printf ("I cannot translate %s element.\n",
                                          "MZ_NAND_OE");
                                 good_bye (TRUE, TRUE);
                                 break;
                        case MZ_OR_OE     :
#if 0
                                 strcpy (topology_vector[te_counter].type, 
                                          "MZ_OR_OE");
#endif
                                 printf ("I cannot translate %s element.\n",
                                          "MZ_OR_OE");
                                 good_bye (TRUE, TRUE);
                                 break;
                        case MZ_NOR_OE    :
#if 0
                                 strcpy (topology_vector[te_counter].type, 
                                          "MZ_NOR_OE");
#endif
                                 printf ("I cannot translate %s element.\n",
                                          "MZ_NOR_OE");
                                 good_bye (TRUE, TRUE);
                                 break;
                        case MZ_BUF_OE    :
#if 0
                                 strcpy (topology_vector[te_counter].type, 
                                          "MZ_BUF_OE");
#endif
                                 printf ("I cannot translate %s element.\n",
                                          "MZ_BUF_OE");
                                 good_bye (TRUE, TRUE);
                                 break;
                        case MZ_INV_OE    :
#if 0
                                 strcpy (topology_vector[te_counter].type, 
                                          "MZ_INV_OE");
#endif
                                 printf ("I cannot translate %s element.\n",
                                          "MZ_INV_OE");
                                 good_bye (TRUE, TRUE);
                                 break;
                        case MZ_SRNOR     :
#if 0
                                 strcpy (topology_vector[te_counter].type, 
                                          "MZ_SRNOR");
#endif
                                 printf ("I cannot translate %s element.\n",
                                          "MZ_SRNOR");
                                 good_bye (TRUE, TRUE);
                                 break;
                        case MZ_MUX       :
#if 0
                                 strcpy (topology_vector[te_counter].type, 
                                          "MZ_MUX");
#endif
                                 printf ("I cannot translate %s element.\n",
                                          "MZ_MUX");
                                 good_bye (TRUE, TRUE);
                                 break;


                        /* flip-flops */

                        case MZ_DFF_PET   :
#if 0
                                 strcpy (topology_vector[te_counter].type, 
                                          "MZ_DFF_PET");
#endif
                                 printf ("I cannot translate %s element.\n",
                                          "MZ_DFF_PET");
                                 good_bye (TRUE, TRUE);
                                 break;

                        case MZ_TFF_PET   :

#if 0
                                 strcpy (topology_vector[te_counter].type, 
                                          "MZ_TFF_PET");
#endif
                                 printf ("I cannot translate %s element.\n",
                                          "MZ_TFF_PET");
                                 good_bye (TRUE, TRUE);
                                 break;

                        case MZ_DFF_PET_C :
#if 0
                                 strcpy (topology_vector[te_counter].type, 
                                          "MZ_DFF_PET_C");
#endif
                                 printf ("I cannot translate %s element.\n",
                                          "MZ_DFF_PET_C");
                                 good_bye (TRUE, TRUE);
                                 break;

                        case MZ_TFF_PET_C :
#if 0
                                 strcpy (topology_vector[te_counter].type, 
                                          "MZ_TFF_PET_C");
#endif
                                 printf ("I cannot translate %s element.\n",
                                          "MZ_TFF_PET_C");
                                 good_bye (TRUE, TRUE);
                                 break;

                        case MZ_DFF_PET_S :
#if 0
                                 strcpy (topology_vector[te_counter].type, 
                                          "MZ_DFF_PET_S");
#endif
                                 printf ("I cannot translate %s element.\n",
                                          "MZ_DFF_PET_S");
                                 good_bye (TRUE, TRUE);
                                 break;

                        case MZ_TFF_PET_S :
#if 0
                                 strcpy (topology_vector[te_counter].type, 
                                          "MZ_TFF_PET_S");
#endif
                                 printf ("I cannot translate %s element.\n",
                                          "MZ_TFF_PET_S");
                                 good_bye (TRUE, TRUE);
                                 break;

                        case MZ_DFF_PET_SC:
#if 0
                                 strcpy (topology_vector[te_counter].type, 
                                          "MZ_DFF_PET_SC");
#endif
                                 printf ("I cannot translate %s element.\n",
                                          "MZ_DFF_PET_SC");
                                 good_bye (TRUE, TRUE);
                                 break;

                        case MZ_TFF_PET_SC :
#if 0
                                 strcpy (topology_vector[te_counter].type, 
                                          "MZ_TFF_PET_SC");
#endif
                                 printf ("I cannot translate %s element.\n",
                                          "MZ_TFF_PET_SC");
                                 good_bye (TRUE, TRUE);
                                 break;

                        case MZ_DFF_NET   :
#if 0
                                 strcpy (topology_vector[te_counter].type, 
                                          "MZ_DFF_NET");
#endif
                                 printf ("I cannot translate %s element.\n",
                                          "MZ_DFF_NET");
                                 good_bye (TRUE, TRUE);
                                 break;

                        case MZ_TFF_NET   :

#if 0
                                 strcpy (topology_vector[te_counter].type, 
                                          "MZ_TFF_NET");
#endif
                                 printf ("I cannot translate %s element.\n",
                                          "MZ_TFF_NET");
                                 good_bye (TRUE, TRUE);
                                 break;

                        case MZ_DFF_NET_C :
#if 0
                                 strcpy (topology_vector[te_counter].type, 
                                          "MZ_DFF_NET_C");
#endif
                                 printf ("I cannot translate %s element.\n",
                                          "MZ_DFF_NET_C");
                                 good_bye (TRUE, TRUE);
                                 break;

                        case MZ_TFF_NET_C :
#if 0
                                 strcpy (topology_vector[te_counter].type, 
                                          "MZ_TFF_NET_C");
#endif
                                 printf ("I cannot translate %s element.\n",
                                          "MZ_TFF_NET_C");
                                 good_bye (TRUE, TRUE);
                                 break;

                        case MZ_DFF_NET_S :
#if 0
                                 strcpy (topology_vector[te_counter].type, 
                                          "MZ_DFF_NET_S");
#endif
                                 printf ("I cannot translate %s element.\n",
                                          "MZ_DFF_NET_S");
                                 good_bye (TRUE, TRUE);
                                 break;

                        case MZ_TFF_NET_S :
#if 0
                                 strcpy (topology_vector[te_counter].type, 
                                          "MZ_TFF_NET_S");
#endif
                                 printf ("I cannot translate %s element.\n",
                                          "MZ_TFF_NET_S");
                                 good_bye (TRUE, TRUE);
                                 break;

                        case MZ_DFF_NET_SC:
#if 0
                                 strcpy (topology_vector[te_counter].type, 
                                          "MZ_DFF_NET_SC");
#endif
                                 printf ("I cannot translate %s element.\n",
                                          "MZ_DFF_NET_SC");
                                 good_bye (TRUE, TRUE);
                                 break;

                        case MZ_TFF_NET_SC :
#if 0
                                 strcpy (topology_vector[te_counter].type, 
                                          "MZ_TFF_NET_SC");
#endif
                                 printf ("I cannot translate %s element.\n",
                                          "MZ_TFF_NET_SC");
                                 good_bye (TRUE, TRUE);
                                 break;

                        case MZ_JKFF_PET    :
#if 0
                                 strcpy (topology_vector[te_counter].type, 
                                          "MZ_JKFF_PET");
#endif
                                 printf ("I cannot translate %s element.\n",
                                          "MZ_JKFF_PET");
                                 good_bye (TRUE, TRUE);
                                 break;

                        case MZ_JKFF_PET_C  :
#if 0
                                 strcpy (topology_vector[te_counter].type, 
                                          "MZ_JKFF_PET_C");
#endif
                                 printf ("I cannot translate %s element.\n",
                                          "MZ_JKFF_PET_C");
                                 good_bye (TRUE, TRUE);
                                 break;

                        case MZ_JKFF_PET_S  :
#if 0
                                 strcpy (topology_vector[te_counter].type, 
                                          "MZ_JKFF_PET_S");
#endif
                                 printf ("I cannot translate %s element.\n",
                                          "MZ_JKFF_PET_S");
                                 good_bye (TRUE, TRUE);
                                 break;

                        case MZ_JKFF_PET_SC :
#if 0
                                 strcpy (topology_vector[te_counter].type, 
                                          "MZ_JKFF_PET_SC");
#endif
                                 printf ("I cannot translate %s element.\n",
                                          "MZ_JKFF_PET_SC");
                                 good_bye (TRUE, TRUE);
                                 break;


                        case MZ_JKFF_NET    :
#if 0
                                 strcpy (topology_vector[te_counter].type, 
                                          "MZ_JKFF_NET");
#endif
                                 printf ("I cannot translate %s element.\n",
                                          "MZ_JKFF_NET");
                                 good_bye (TRUE, TRUE);
                                 break;

                        case MZ_JKFF_NET_C  :
#if 0
                                 strcpy (topology_vector[te_counter].type, 
                                          "MZ_JKFF_NET_C");
#endif
                                 printf ("I cannot translate %s element.\n",
                                          "MZ_JKFF_NET_C");
                                 good_bye (TRUE, TRUE);
                                 break;

                        case MZ_JKFF_NET_S  :
#if 0
                                 strcpy (topology_vector[te_counter].type, 
                                          "MZ_JKFF_NET_S");
#endif
                                 printf ("I cannot translate %s element.\n",
                                          "MZ_JKFF_NET_S");
                                 good_bye (TRUE, TRUE);
                                 break;

                        case MZ_JKFF_NET_SC :
#if 0
                                 strcpy (topology_vector[te_counter].type, 
                                          "MZ_JKFF_NET_SC");
#endif
                                 printf ("I cannot translate %s element.\n",
                                          "MZ_JKFF_NET_SC");
                                 good_bye (TRUE, TRUE);
                                 break;

                        case MZ_LATCH_HIGH  :
#if 0
                                 strcpy (topology_vector[te_counter].type, 
                                          "MZ_LATCH_HIGH");
#endif
                                 printf ("I cannot translate %s element.\n",
                                          "MZ_LATCH_HIGH");
                                 good_bye (TRUE, TRUE);
                                 break;

                        case MZ_LATCH_LOW  :
#if 0
                                 strcpy (topology_vector[te_counter].type, 
                                          "MZ_LATCH_LOW");
#endif
                                 printf ("I cannot translate %s element.\n",
                                          "MZ_LATCH_LOW");
                                 good_bye (TRUE, TRUE);
                                 break;

                        default: printf ("I found an unknown gate type.\n");
                                 good_bye (TRUE, TRUE);
                     }

                     break;

            default: printf 
                        ("Some elements in the network are not supported.\n");
                     good_bye (TRUE, TRUE);

         }  /* end of switch (te_code) */



         /* fanout list */

         for (link_counter = 0; link_counter < te_fo_num; link_counter++)
         {
            fscanf (net_file, "%ld %ld", &link_type, &link_index);

            fanout_vector[fo_offset++].index = link_index;

            if (te_code == MZ_RT_ASSIGNEMENT)
            {
               /* for now, ignore the information */

               fscanf (net_file, "%ld %ld %ld %ld", &link_b1c, &link_b2c,
                                                    &link_b1d, &link_b2d);
            }
            else
            {
               /* for now, ignore the information */

               fscanf (net_file, "%ld %ld", &link_pin, &link_item);
            }
        }


         /* fanin list */

         for (link_counter = 0; link_counter < te_fi_num; link_counter++)
         {
            fscanf (net_file, "%ld %ld %ld %ld", &link_type, &link_index,
                                                 &link_bit, &link_item);

            fanin_vector[fi_offset++].index = link_index;

            if (link_bit != 0)
            {
               printf ("The bit extraction is not supported.\n");
               good_bye (TRUE, TRUE);
            }
         }


         /* connectivity list */

#if 0

         for (link_counter = 0; link_counter < te_conn_num; link_counter++)
         {
            fscanf (net_file, "%ld %ld %ld %ld", &link_type, &link_index,
                                                 &link_bit, &link_item);

            if (link_bit != 0)
            {
               printf ("The bit extraction is not supported.\n");
               good_bye (TRUE, TRUE);
            }

         }

#endif


         /* the computational appendix for now cannot be loaded */

         if ((te_code == MZ_RT_ASSIGNEMENT) || (te_code == MZ_RT_CONDITION))
         {
            printf ("The RT assignements and conditions are not supported.\n");
            good_bye (TRUE, TRUE);
         }

      }  /* end of one element loading */

   }  /* end of all the elements loading */


   /* the initialization section is not supported */

   if (init_num != 0)
   {
      printf 
("Warning: the initialization of the elements is ignored (see output file).\n");

      fprintf (hilo_file, "\n**\n** Initializations:\n**\n");
   }

   {

      /* local variables */

      unsigned long  init_values_counter;
      unsigned long  init_element_counter;
      unsigned char  init_string [MAX_DATA_SIZE];
      unsigned long  init_index;


      /* beginning of executable part */

      for (init_values_counter = 0; init_values_counter < init_num;
                                                       init_values_counter++)
      {
         /* read the initialization value and the number of elements
            to be initialized */

         fscanf (net_file, "%s %ld", init_string, &init_element_counter);

         for ( ; init_element_counter > 0; init_element_counter--)
         {
            /* read the index of the element to be initialized.
               The frame information is not yet used */

            fscanf (net_file, "%ld", &init_index);

            fprintf (hilo_file, "** %s = %s\n", 
                     topology_vector[init_index].name, init_string);
         }
      }
   }


   printf ("Network loaded.\n");


   return;

}  /* end of load_topology */



/******************************************************************************

    Function name:    translate

    Authors:          Pier Luca Montessoro

    Creation date:    21 June 1988

    Purpose:          to translate the .dee file into hilo source

 ******************************************************************************/


translate (comp_delay)

unsigned char  comp_delay;

{

   /* local variables */

   unsigned char         comma;
   unsigned char         str_offset [80];
   unsigned long         curr_column;
   unsigned long         start_column;
   unsigned long         index_cnt;
   unsigned long         link_cnt;
   unsigned long         link_offset;
   unsigned long         rise_delay;
   unsigned long         fall_delay;
   unsigned long         rise_del_coeff;
   unsigned long         fall_del_coeff;


   /* beginning of executable part */

   /* write the header */

   fprintf (hilo_file, "\nCCT %s (", experiment_name);

   start_column = strlen (experiment_name) + 6;

   for (curr_column = 0; curr_column < start_column; curr_column++)
   {
      str_offset [curr_column] = ' ';
   }
   str_offset [curr_column] = '\0';

   comma = FALSE;

   for (index_cnt = 0; index_cnt < elements_num; index_cnt++)
   {
      if (topology_vector[index_cnt].it_is_in ||  
          topology_vector[index_cnt].it_is_out)
      {
         if ((curr_column = 
                 curr_column + strlen (topology_vector[index_cnt].name) + 2) 
                                                                         >= 80)
         {
            if (comma)
            {
               fprintf (hilo_file, ",\n%s%s", str_offset, 
                        topology_vector[index_cnt].name);
            }
            else
            {
               fprintf (hilo_file, "\n%s", topology_vector[index_cnt].name);
               comma  = TRUE;
            }
            curr_column = start_column + 
                          strlen (topology_vector[index_cnt].name);
         }
         else
         {
            if (comma)
            {
               fprintf (hilo_file, ", %s", topology_vector[index_cnt].name);
            }
            else
            {
               fprintf (hilo_file, "%s", topology_vector[index_cnt].name);
               comma  = TRUE;
            }
         }
      }
   }

   fprintf (hilo_file, ")\n\n");


   /* write the element instances */

   if (comp_delay)
   {
      rise_del_coeff = RISE_DEL_COEFF;
      fall_del_coeff = FALL_DEL_COEFF;
   }
   else
   {
      rise_del_coeff = 0;
      fall_del_coeff = 0;
   }


   for (index_cnt = 0; index_cnt < elements_num; index_cnt++)
   {
      if (!topology_vector[index_cnt].it_is_only_a_wire)
      {
         /* type, delay, instance name */

         rise_delay = 
            delay_table[topology_vector[index_cnt].delay_index].rise + 
            rise_del_coeff * topology_vector[index_cnt].fanout_number;

         fall_delay = 
            delay_table[topology_vector[index_cnt].delay_index].fall + 
            fall_del_coeff * topology_vector[index_cnt].fanout_number;

         if (max_r_del < rise_delay)
         {
            max_r_del = rise_delay;
         }

         if (max_f_del < fall_delay)
         {
            max_f_del = fall_delay;
         }


         fprintf (hilo_file, "%s (%ld, %ld) ge%ld (", 
                  topology_vector[index_cnt].type, rise_delay, fall_delay, 
                  index_cnt);


         /* out */

         /* if the element writes on a wire, that name must be the name 
            of the output wire */

         if (topology_vector[index_cnt].fanout_number > 0)
         {
            if (topology_vector[
                   fanout_vector[
                      topology_vector[
                         index_cnt]
                      .fanout_position]
                   .index]
                .it_is_only_a_wire)
            {
               /* be sure that it is the only output */
               if (topology_vector[index_cnt].fanout_number != 1)
               {
                  printf 
 ("I don't understand: an elements writes on a wire AND on something else!\n");
                  good_bye (TRUE, TRUE);
               }

               fprintf (hilo_file, "%s, ", topology_vector[
                                              fanout_vector[
                                                 topology_vector[
                                                    index_cnt]
                                                 .fanout_position]
                                              .index]
                                           .name);
            }
            else
            {
               fprintf (hilo_file, "%s, ", topology_vector[index_cnt].name);
            }
         }
         else
         {
            fprintf (hilo_file, "%s, ", topology_vector[index_cnt].name);
         }


         /* in */

         for (link_cnt = 0, 
                 link_offset = topology_vector[index_cnt].fanin_position;
              link_cnt < topology_vector[index_cnt].fanin_number - 1; 
              link_cnt++, link_offset++)
         {
            fprintf (hilo_file, "%s, ", 
                     topology_vector[fanin_vector[link_offset].index].name);
         }
         fprintf (hilo_file, "%s);\n", 
                  topology_vector[fanin_vector[link_offset].index].name);

      }
   }



   /* write the wires declaration */

   fprintf (hilo_file, "\nWIRE ");
   curr_column = 5;

   for (index_cnt = 0; index_cnt < elements_num; index_cnt++)
   {
      if (!(topology_vector[index_cnt].it_is_only_a_wire || 
            topology_vector[index_cnt].it_is_in)) 
      {
         if ((curr_column = 
                 curr_column + strlen (topology_vector[index_cnt].name) + 2) 
                                                                         >= 80)
         {
            fprintf (hilo_file, "\n      %s", topology_vector[index_cnt].name);
            curr_column = 5 + strlen (topology_vector[index_cnt].name);
         }
         else
         {
            fprintf (hilo_file, " %s", topology_vector[index_cnt].name);
         }
      }
   }
   fprintf (hilo_file, ";\n");


   /* declare vdd and gnd, if present */

   for (index_cnt = 0; index_cnt < elements_num; index_cnt++)
   {
      if (!strcmp (topology_vector[index_cnt].name, "VDD"))
      {
         fprintf (hilo_file, "\nSUPPLY1  VDD;\n");
      }
      else
      {
         if (!strcmp (topology_vector[index_cnt].name, "GND"))
         {
            fprintf (hilo_file, "\nSUPPLY0  GND;\n");
         }
      }
   }


   /* write the inputs declaration */

   fprintf (hilo_file, "\nINPUT ");
   curr_column = 6;

   for (index_cnt = 0; index_cnt < elements_num; index_cnt++)
   {
      if (topology_vector[index_cnt].it_is_in)
      {
         if ((curr_column = 
                 curr_column + strlen (topology_vector[index_cnt].name) + 1) 
                                                                         >= 80)
         {
            fprintf (hilo_file, "\n       %s", 
                     topology_vector[index_cnt].name);
            curr_column = 6 + strlen (topology_vector[index_cnt].name);
         }
         else
         {
            fprintf (hilo_file, " %s", topology_vector[index_cnt].name);
         }
      }
   }
   fprintf (hilo_file, ".\n\n");

   printf ("Network translated.\n");
   printf ("Maximum delays: rise %ld, fall %ld.\n", max_r_del, max_f_del);


   return;

}  /* end of translate */



/******************************************************************************

    Function name:    good_bye

    Authors:          Pier Luca Montessoro

    Creation date:    21 June 1988

    Purpose:          exit function

 ******************************************************************************/


good_bye (open_files, error_found)

unsigned char  open_files;
unsigned char  error_found;

{

   /* beginning of executable part */

   if (open_files)
   {
      fclose (net_file);
      fclose (hilo_file);
   }

   if (error_found)
   {
      printf ("I am sorry, but I cannot proceed. Try again.\n");
   }
   else
   {
      printf ("Job terminated. Good-bye!");
   }

   exit (1);

}  /* end of good_bye */



/******************************************************************************

    Function name:    get_memory

    Authors:          Pier Luca Montessoro

    Creation date:    25 January 1988

    Purpose:          memory allocation for network data structures

 ******************************************************************************/


char  *get_memory (number, size)

unsigned long  number;
unsigned long  size;

{

   /* local variable */

   char   *alloc_pointer;


   /* beginning of executable part */

   alloc_pointer = (char *) malloc (number * size);


   if (alloc_pointer == NULL)
   {
      printf ("No memory is available for topological data allocation.\n");
      good_bye (TRUE, TRUE);
   }

   return (alloc_pointer);

}  /* end of get_memory */



/******************************************************************************

    Function name:    load_sym_tab

    Authors:          Pier Luca Montessoro

    Creation date:    12 April 1988

    Purpose:          to load the symbol table of the current experiment

 ******************************************************************************/


load_sym_tab ()

{

   /* local variables */

   int                   saved_byte_offset;
   char                  *st_offset;
   char                  *el_name;
   char                  dummy_str [MAX_NAME_LENGTH];
   unsigned long         dummy_num;                                        
   unsigned long         mod_size;
   unsigned long         n_modules;
   unsigned long         it_size;
   unsigned long         t_index;
   unsigned long         i;
   unsigned long         j;
   unsigned char         us_converted;


   /* beginning of executable part */

   /* save the current position in the net file */

   saved_byte_offset = ftell (net_file);

   rewind (net_file);

   /* read the address of the first record (the first and second lines are 
      skipped) */

   while (getc (net_file) != '\n');
   while (getc (net_file) != '\n');

   fscanf (net_file, "%ld", &st_offset);   


   /* position the file pointer at the beginning of the symbol table */

   if (fseek (net_file, st_offset, 0))
   {
      printf ("Reading symbol table failed.\n");
      good_bye (TRUE, TRUE);
   }


   /* read the number of modules */
   fscanf (net_file, "%ld %ld", &n_modules, &dummy_num);


   /* reach the main module (the first one) */

   for (j = 0; j < n_modules; j++)
   {

      /* read the size of the interface table */   
      fscanf (net_file, "%s %ld %ld %s %ld %ld %ld %ld", 
              dummy_str, &dummy_num, &dummy_num, dummy_str, 
              &dummy_num, &dummy_num, &dummy_num, &it_size);

      /* skip the interface table (it should be empty for the main module) */
      for (i = 0; i < it_size; i++)
      {
         fscanf (net_file, "%ld %s", &dummy_num, dummy_str);
      }
   }


   /* read the module name and the size */
   fscanf (net_file, "%s %ld", dummy_str, &mod_size);

   us_converted = FALSE;

   /* load the names */
   for (i = 0; i < mod_size; i++)
   {
      fscanf (net_file, "%ld %ld %s", &t_index, &dummy_num, dummy_str);

      /* load the name only if belonging to a wire */

      if (dummy_num == MZ_WIRE)
      {
         /* convert the name to uppercase */
         el_name = topology_vector[t_index].name;
         for (j = 0; dummy_str[j] != '\0'; j++)
         {
            if (dummy_str[j] == '_')
            {
               us_converted = TRUE;
               el_name[j] = 'X';
            }
            else
            {
               el_name[j] = toupper (dummy_str[j]);
            }
         }
         el_name[j] = '\0';
      }
   }

   fseek (net_file, saved_byte_offset, 0);

   printf ("Symbol table loaded.\n");

   if (us_converted)
   {
      printf ("Warning: some '_' in names have been converted to 'X'.\n");
   }

   return;

}  /* end of load_sym_tab */
